/*
 * atapi.c - ATAPI support
 */
#include "global.h"

#ifdef SUPPORT_ATAPI

#define CD_FRAMESIZE       2048
#define CD_FRAMESHIFT      11
#define SECTOR_SIZE	       512
#define SECTORS_PER_FRAME  (CD_FRAMESIZE / SECTOR_SIZE)

BOOLEAN fAtapiInterrupt(_VBUS_ARG PDevice pDev)
{
	PCommand pCmd = pDev->pChannel->CurrentCmd;

	if (pCmd && pDev->Handler) {
		pCmd->Result = pDev->Handler(_VBUS_P pDev);
		if (pCmd->Result!=RETURN_PENDING) {
			if(pDev->pChannel->cf_do_dma == 0)
				pDev->pChannel->pChipInstance->ftbl.pfnModeSetting(pDev->pChannel, pDev,
										  pDev->bDeModeSetting);
			CallAfterReturn(_VBUS_P (DPC_PROC)pCmd->pfnCompletion, pCmd);
		}
	}
	else
		HPT_ASSERT(0);
	return TRUE;
}

static void atapi_fail(PDevice pDev)
{
	PVDevice pVDevice = Map2pVDevice(pDev);
	_VBUS_INST(pVDevice->pVBus)

	pDev->df_on_line = 0;
	pVDevice->vf_online = 0;

	if (pVDevice->pfnDeviceFailed) 
		CallWhenIdle(_VBUS_P (DPC_PROC)pVDevice->pfnDeviceFailed, pVDevice);
	fNotifyGUI(ET_DEVICE_REMOVED, pVDevice);
}

static UINT start_packet_command(_VBUS_ARG PDevice pDev, UINT xferlen)
{
    PChannel          pChan = pDev->pChannel;
    PIDE_REGISTERS_1  IoPort = pChan->BaseIoAddress1;
    PIDE_REGISTERS_2  ControlPort = pChan->BaseIoAddress2;
	UCHAR status;

	status = pChan->pChipInstance->ftbl.pfnWaitOnBusy(pChan, pDev->bDeUnitId, TRUE);
	if (status & IDE_STATUS_BUSY)
		return RETURN_DEVICE_BUSY;

	if (status & IDE_STATUS_DWF){
		atapi_fail(pDev);
	    return RETURN_BAD_DEVICE;
	}

	mSetFeaturePort(IoPort, (UCHAR)((pChan->cf_do_dma)? 1 : 0));
	mSetBlockCount(IoPort, 0);
	mSetBlockNumber(IoPort, 0);
	mSetCylinderLow(IoPort, (UCHAR)(xferlen & 0xff));
	mSetCylinderHigh(IoPort, (UCHAR)(xferlen>>8));

	mIssueCommand(IoPort, 0xA0);
 
	if(pDev->df_multiple==0)
		CallAfterReturn(_VBUS_P (DPC_PROC)fAtapiInterrupt, pDev);
	return RETURN_PENDING;
}

static UINT check_error(_VBUS_ARG PDevice pDev)
{
	PIDE_REGISTERS_1  IoPort = pDev->pChannel->BaseIoAddress1;
	PCommand pCmd = pDev->pChannel->CurrentCmd;
	UCHAR err, sense_key;

	err = mGetErrorCode(IoPort);
	sense_key = err >> 4;

    KdPrint(("sense_key:%x",sense_key));

	if (pCmd->cf_atapi==0 ||
		pCmd->uCmd.Atapi[0]==ATAPI_READ ||
		pCmd->uCmd.Atapi[0]==ATAPI_WRITE)
	{
		/* Handle errors from READ and WRITE requests. */
		if (sense_key == ATAPI_SENSE_NOT_READY) {
			/* Tray open. */
			pDev->df_media_changed = 1;
			return RETURN_IDE_ERROR;
		} 
		else if (sense_key == ATAPI_SENSE_UNIT_ATTENTION) {
			/* Media change. */
			pDev->df_media_changed = 1;
			/*return RETURN_RETRY; */
		}
		else if (sense_key == ATAPI_SENSE_ILLEGAL_REQUEST ||  
			sense_key == ATAPI_SENSE_DATA_PROTECT) {
			return RETURN_INVALID_REQUEST;
		}

		/* If we got a CHECK_CONDITION status, queue a request sense command. */
		/*if (mGetBaseStatus(IoPort) & IDE_STATUS_ERROR) */
		/*	queue_request_sense(_VBUS_P pDev); */

		return RETURN_IDE_ERROR;
	}
	else if (pCmd->uCmd.Atapi[0]==ATAPI_REQUEST_SENSE) {
		/* We got an error trying to get sense info
		   from the drive (probably while trying
		   to recover from a former error).  Just give up. */
		return RETURN_IDE_ERROR;

	} else {
		/* All other functions, except for READ. */
		/* Check for tray open. */
		if (sense_key == ATAPI_SENSE_NOT_READY) {
			pDev->df_media_changed = 1;
			return RETURN_IDE_ERROR;
		}
		else if (sense_key == ATAPI_SENSE_UNIT_ATTENTION) {
			/* Check for media change. */
			pDev->df_media_changed = 1;
			/*return RETURN_PENDING; */
		}
		else{
			/* Otherwise, print an error. */
			KdPrint(("packet cmd error (%d)", sense_key));
		}
		
		/*if (mGetBaseStatus(IoPort) & IDE_STATUS_ERROR) */
		/*	queue_request_sense(_VBUS_P pDev); */
	}
	return RETURN_IDE_ERROR;
}

static UINT transfer_packet_command(_VBUS_ARG PDevice pDev)
{
	PChannel          pChan = pDev->pChannel;
	PIDE_REGISTERS_1  IoPort = pChan->BaseIoAddress1;
	PIDE_REGISTERS_2  ControlPort = pChan->BaseIoAddress2;
    PUCHAR            BMI = pChan->BMI;
	UCHAR             status;

	if(pDev->df_multiple) {
		/* Check for errors. */
		status = mGetBaseStatus(IoPort);
		if (status & IDE_STATUS_DWF)
			goto bad_device;
		if ((status & IDE_STATUS_DRQ)==0)
			return check_error(_VBUS_P pDev);
	} else {
		status = fWaitOnBaseBusy(IoPort);
		if (status & IDE_STATUS_DWF){
bad_device:
			atapi_fail(pDev);
			return RETURN_BAD_DEVICE;
		}
		if ((fWaitForDrq(ControlPort) & IDE_STATUS_DRQ)==0)
			return RETURN_WRITE_NO_DRQ;
	}
	if(pChan->cf_do_dma)
	{
		OutPort(BMI, BMI_CMD_STOP);
	    OutPort(BMI + BMI_STS, BMI_STS_INTR);  
	}

	RepOUTS(IoPort, (void *)pDev->pc, 6);

	if(pChan->cf_do_dma)
	{
		OutDWord(BMI + BMI_DTP , pChan->SgPhysicalAddr);
		OutPort(BMI, (UCHAR)(pChan->CurrentCmd->cf_data_in? BMI_CMD_STARTREAD : BMI_CMD_STARTWRITE));
	}
	return RETURN_PENDING;
}

#define IF_I960(x)
#define IIF_I960(x) if (0)
static UINT cdrom_pio_intr(_VBUS_ARG PDevice pDev)
{
	PIDE_REGISTERS_1  IoPort = pDev->pChannel->BaseIoAddress1;
	PCommand pCmd = pDev->pChannel->CurrentCmd;
	UCHAR status, ireason;
	UINT len;
	ULONG count;

	status = fWaitOnBaseBusy(IoPort);
	if (status & IDE_STATUS_DWF){
		atapi_fail(pDev);
		return RETURN_BAD_DEVICE;
	}

	if (status & IDE_STATUS_ERROR) {
		len = check_error(_VBUS_P pDev);
		if (len!=RETURN_PENDING)
			return len;
	}

	/* Read the interrupt reason and the transfer length. */
	ireason = mGetInterruptReason(IoPort);
	len = mGetByteLow(IoPort) | (mGetByteHigh(IoPort)<<8);
	len>>=1;

	/* If DRQ is clear, the command has completed. */
	if ((status & IDE_STATUS_DRQ)==0) {
		if (pCmd->uScratch.disk.WordsLeft) {
			/* some commands' transfer length may be less than allocation length */
			if (pCmd->cf_atapi) {
				switch(pCmd->uCmd.Atapi[0]) {
				case ATAPI_REQUEST_SENSE:
				case ATAPI_INQUIRY:
				case ATAPI_MODE_SENSE10:
				case ATAPI_READ_HEADER:
				case ATAPI_READ_TOC:
				case ATAPI_READ_SUB_CHANNEL:
				case ATAPI_GET_CONFIGURATION:
				case ATAPI_READ_DISK_INFORMATION:
				case ATAPI_READ_TRACK_INFORMATION:
					return RETURN_SUCCESS;
				}
			}
			KdPrint(("data underrun"));
			return RETURN_IDE_ERROR;
		}
		else
			return RETURN_SUCCESS;
	}

	ireason &= 3;

	switch(ireason) {
	case 2:
		if (pCmd->cf_data_out) {
			/* the drive wants to send data. */
			while (len--) InWord(IoPort);
			return RETURN_IDE_ERROR;
		}
		break;
	case 1:
		/* Some drives (ASUS) seem to tell us that status
		 * info is available. just get it and ignore.
		 */
		mGetBaseStatus(IoPort);
		break;
	case 0:
		if (pCmd->cf_data_in) {
			/* the drive is expecting to receive data from us! */
			KdPrint(("Drive wants to transfer data the wrong way"));
			/* Throw some data at the drive so it doesn't hang and quit this request. */
			while (len--) OutWord(IoPort, 0);
			return RETURN_IDE_ERROR;
		}
		break;
	default:
		/* Drive wants a command packet, or invalid ireason... */
		KdPrint(("bad interrupt reason %d", ireason));
		return RETURN_IDE_ERROR;
	}

	/* ship data if needed */
	if (pCmd->cf_data_in) {
		count = pCmd->uScratch.disk.WordsSkip;
		if (count>len) count = len;
		pCmd->uScratch.disk.WordsSkip -= count;
		len -= count;
		while (count--) InWord(IoPort);
	}

	if (len) {
		UINT  SgWords, ThisWords;
		FPSCAT_GATH pSG = (FPSCAT_GATH)pCmd->uScratch.disk.pPIOSg;

		if (len>pCmd->uScratch.disk.WordsLeft) {
			count = len - pCmd->uScratch.disk.WordsLeft;
			len = pCmd->uScratch.disk.WordsLeft;
		}
		else
			count = 0;

		pCmd->uScratch.disk.WordsLeft -= len;

		while(len > 0) {
			IF_I960 (UINT onebyte);
    		if((SgWords = pSG->wSgSize) == 0) {
    			IF_I960(onebyte = 0);
				SgWords = 0x8000;
      		}
      		else {
      			IF_I960 (onebyte = SgWords & 1);
				SgWords >>= 1;
			}

    		ThisWords = MIN(SgWords, (UINT)len);

    		if(pCmd->cf_data_out)
        		RepOUTS(IoPort, (ADDRESS)pSG->dSgAddress, ThisWords);
    		else
        		RepINS(IoPort, (ADDRESS)pSG->dSgAddress, ThisWords);

    		if((SgWords -= (USHORT)ThisWords) == 0) {
    			IIF_I960(onebyte) {
    				WORD w = InWord(IoPort);
    				((BYTE FAR *)pSG->dSgAddress)[ThisWords<<1] = (BYTE)w;
    				pSG++;
    				((BYTE FAR *)pSG->dSgAddress)[0] = (BYTE)(w>>8);
    				pSG->dSgAddress++;
    				pSG->wSgSize--;
    				len -= ThisWords+1;
    			}
    			else {
        			len -= ThisWords;
        			pSG++;
        		}
    		}
    		else {
        		pSG->dSgAddress += (ThisWords * 2);
           		pSG->wSgSize -= (USHORT)(ThisWords * 2);
        		break;
    		}
		}
		if (pCmd->cf_data_in) {
			/* drop extra data */
			while (count--) InWord(IoPort);
		}
		pCmd->uScratch.disk.pPIOSg = pSG;
	}
	pDev->Handler = cdrom_pio_intr;
	return RETURN_PENDING;
}

static UINT start_rw_cmd(_VBUS_ARG PDevice pDev)
{
	PCommand pCmd = pDev->pChannel->CurrentCmd;
	LBA_T frame, Lba = pCmd->uCmd.Ide.Lba;
	USHORT nframes, nskip, nSectors = pCmd->uCmd.Ide.nSectors;

	/*
	 * If the requested sector doesn't start on a cdrom block boundary,
	 * we must adjust the start of the transfer so that it does,
	 * and remember to skip the first few sectors.
	 */
	nskip = (USHORT)(Lba % SECTORS_PER_FRAME);
	if (nskip) {
		Lba -= nskip;
		nSectors += nskip;
	}
	pCmd->uScratch.disk.WordsSkip = (ULONG)nskip<<8;

	/* 
	 * convert from sectors to cdrom blocks, rounding up the transfer
	 * length if needed. 
	 */
	nframes = (nSectors + SECTORS_PER_FRAME-1) / SECTORS_PER_FRAME;
	frame = Lba / SECTORS_PER_FRAME;

	/* Set up the command */
	memset (pDev->pc, 0, 12);
	pDev->pc[0] = ATAPI_READ;
	pDev->pc[2] = (UCHAR)(frame>>24);
	pDev->pc[3] = (UCHAR)(frame>>16);
	pDev->pc[4] = (UCHAR)(frame>>8);
	pDev->pc[5] = (UCHAR)(frame & 0xff);
	pDev->pc[7] = (UCHAR)(nframes >> 8);
	pDev->pc[8] = (UCHAR)(nframes & 0xff);

	pDev->Handler = cdrom_pio_intr;
	return transfer_packet_command(_VBUS_P pDev);
}

static UINT start_atapi_cmd(_VBUS_ARG PDevice pDev)
{
	PCommand pCmd = pDev->pChannel->CurrentCmd;
	memcpy(pDev->pc, pCmd->uCmd.Atapi, 12);
	pDev->Handler = cdrom_pio_intr;
	return transfer_packet_command(_VBUS_P pDev);
}

static ULONG atapi_len(PUCHAR pc)
{
	switch(pc[0]) {
		case ATAPI_READ:
	case ATAPI_WRITE:
	case ATAPI_VERIFY:
		return (((ULONG)pc[7]<<8)|pc[8]) << CD_FRAMESHIFT;
	case ATAPI_READ12:
		return (((ULONG)pc[6]<<24)|((ULONG)pc[7]<<16)|((ULONG)pc[8]<<8)|pc[9]) << CD_FRAMESHIFT;
	case ATAPI_MODE_SELECT10:
	case ATAPI_MODE_SENSE10:
	case ATAPI_READ_HEADER:
	case ATAPI_READ_SUB_CHANNEL:
	case ATAPI_READ_BUFFER_CAPACITY:
	case ATAPI_GET_CONFIGURATION:
	case ATAPI_GET_EVENT_STATUS_NOTIFICATION:
	case ATAPI_READ_DATA_BUFF:    /* need fix */
		return (((ULONG)pc[7]<<8)|pc[8]);
	case ATAPI_READ_CAPACITY:
		return 8;
	case ATAPI_REQUEST_SENSE:
	case ATAPI_INQUIRY:
		return pc[4];
	case ATAPI_REPORT_KEY:
	case ATAPI_READ_DVD_STRUCTURE:
	case ATAPI_MECHANISM_STATUS:
		return (((ULONG)pc[8]<<8)|pc[9]);
	case ATAPI_READ_CD:
		/* not for sub-channel */
		/* return LongMul((((ULONG)pc[6]<<16)|((ULONG)pc[7]<<8)|pc[8]), 2352); */
	case ATAPI_SET_CDRW_SPEED:
	case ATAPI_READ_CD_MSF:
	case ATAPI_READ_TOC:
	case ATAPI_READ_DISK_INFORMATION:
	case ATAPI_READ_TRACK_INFORMATION: 
		return (ULONG)-1;
	/* unsupported commands */
	case ATAPI_MODE_SELECT:
	case ATAPI_MODE_SENSE:
		HPT_ASSERT(0);
	default:
		return 0;
	}
}

void fStartAtapiCommand(_VBUS_ARG PDevice pDev, PCommand pCmd)
{
    PChannel pChan = pDev->pChannel;
    int DO_DMA = 0;

	if (pCmd->cf_atapi==0) {
		/* IDE command format. must be read/write */
		if (!pCmd->cf_data_in && !pCmd->cf_data_out) {
invalid_req:
	        KdPrint(("Invalid Cmd!"));
			pCmd->Result = RETURN_INVALID_REQUEST;
			CallAfterReturn(_VBUS_P (DPC_PROC)pCmd->pfnCompletion, pCmd);
			return;
		}
		pDev->Handler = start_rw_cmd;
		if (pDev->df_dma && (pCmd->uCmd.Ide.Lba & 3)==0 && (pCmd->uCmd.Ide.nSectors & 3)==0)
				DO_DMA = 1;
	}
	else
	{
		pDev->Handler = start_atapi_cmd;
		if (pDev->df_dma &&
			((pCmd->uCmd.Atapi[0] == ATAPI_READ) || (pCmd->uCmd.Atapi[0] == ATAPI_WRITE) ||
			  (pCmd->uCmd.Atapi[0] == ATAPI_READ12)))
				DO_DMA = 1;
	}

	pCmd->uScratch.disk.WordsLeft = 0;

	if (pCmd->cf_data_in || pCmd->cf_data_out) {
		if(DO_DMA == 0) 
		{
pio:
			if (pCmd->pSgTable && pCmd->cf_physical_sg==0) {
				FPSCAT_GATH sg1=pChan->pSgTable, sg2=pCmd->pSgTable;
				do { *sg1++=*sg2; } while ((sg2++->wSgFlag & SG_FLAG_EOT)==0);
			}
			else {
				if (!pCmd->pfnBuildSgl || !pCmd->pfnBuildSgl(_VBUS_P pCmd, pChan->pSgTable, 1)) {
					pCmd->Result = RETURN_NEED_LOGICAL_SG;
					CallAfterReturn(_VBUS_P (DPC_PROC)pCmd->pfnCompletion, pCmd);
					return;
				}
			}
			if (pCmd->cf_atapi) {
				ULONG sg_len=0, len = atapi_len(pCmd->uCmd.Atapi);
				FPSCAT_GATH pSg = pChan->pSgTable;
				do {
					sg_len += (pSg->wSgSize? pSg->wSgSize : 0x10000); 
				}
				while ((pSg++->wSgFlag & SG_FLAG_EOT)==0);
				if (len==(ULONG)-1)
					len = sg_len;
				else if (len>sg_len)
					goto invalid_req;
				else
                    len = sg_len;
				pCmd->uScratch.disk.WordsLeft = len>>1;
			}
			else
				pCmd->uScratch.disk.WordsLeft = pCmd->uCmd.Ide.nSectors<<8;
			pCmd->uScratch.disk.pPIOSg = pChan->pSgTable;
		}
		else
		{
			if (pCmd->pSgTable && pCmd->cf_physical_sg) {
				FPSCAT_GATH sg1=pChan->pSgTable, sg2=pCmd->pSgTable;
				do { *sg1++=*sg2; } while ((sg2++->wSgFlag & SG_FLAG_EOT)==0);
			}
			else {
				if (!pCmd->pfnBuildSgl || !pCmd->pfnBuildSgl(_VBUS_P pCmd, pChan->pSgTable, 0))
					goto pio;
			}

			pChan->cf_do_dma = 1;
		}
	}

    if(pChan->cf_do_dma == 0)
		pChan->pChipInstance->ftbl.pfnModeSetting(pChan, pDev, pDev->bDeBestPIO);

	if ((pCmd->Result = start_packet_command(_VBUS_P pDev, 0x8000))) {
		CallAfterReturn(_VBUS_P (DPC_PROC)pCmd->pfnCompletion, pCmd);
	}
}

#endif
